from __future__ import absolute_import

from spidermon.exceptions import NotConfigured
from scrapy.mail import MailSender

from . import SendEmail


DEFAULT_SMTP_ENFORCE_TLS = False
DEFAULT_SMTP_ENFORCE_SSL = False
DEFAULT_SMTP_PORT = 25


class SendSmtpEmail(SendEmail):
    def __init__(
        self,
        smtp_host=None,
        smtp_port=None,
        smtp_user=None,
        smtp_password=None,
        smtp_enforce_tls=None,
        smtp_enforce_ssl=None,
        *args,
        **kwargs
    ):
        super(SendSmtpEmail, self).__init__(*args, **kwargs)

        self.smtp_host = smtp_host
        self.smtp_user = smtp_user
        self.smtp_password = smtp_password
        self.smtp_port = smtp_port or DEFAULT_SMTP_PORT
        self.smtp_enforce_tls = (
            smtp_enforce_tls
            if smtp_enforce_tls is not None
            else DEFAULT_SMTP_ENFORCE_TLS
        )
        self.smtp_enforce_ssl = (
            smtp_enforce_ssl
            if smtp_enforce_ssl is not None
            else DEFAULT_SMTP_ENFORCE_SSL
        )

        if not self.smtp_host:
            raise NotConfigured(
                "You must provide a value for SPIDERMON_SMTP_HOST setting."
            )
        if not self.smtp_user:
            raise NotConfigured(
                "You must provide a value for SPIDERMON_SMTP_USER setting."
            )
        if not self.smtp_password:
            raise NotConfigured(
                "You must provide a value for SPIDERMON_SMTP_PASSWORD setting."
            )

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super(SendSmtpEmail, cls).from_crawler_kwargs(crawler)
        kwargs.update(
            {
                "smtp_host": crawler.settings.get("SPIDERMON_SMTP_HOST"),
                "smtp_port": crawler.settings.getint("SPIDERMON_SMTP_PORT") or None,
                "smtp_user": crawler.settings.get("SPIDERMON_SMTP_USER"),
                "smtp_password": crawler.settings.get("SPIDERMON_SMTP_PASSWORD"),
                "smtp_enforce_tls": crawler.settings.getbool(
                    "SPIDERMON_SMTP_ENFORCE_TLS"
                )
                if "SPIDERMON_SMTP_ENFORCE_TLS" in crawler.settings
                else None,
                "smtp_enforce_ssl": crawler.settings.getbool(
                    "SPIDERMON_SMTP_ENFORCE_SSL"
                )
                if "SPIDERMON_SMTP_ENFORCE_SSL" in crawler.settings
                else None,
            }
        )
        return kwargs

    def send_message(self, message, **kwargs):
        mail_sender = MailSender(
            smtphost=self.smtp_host,
            mailfrom=self.sender,
            smtpuser=self.smtp_user,
            smtppass=self.smtp_password,
            smtpport=self.smtp_port,
            smtptls=self.smtp_enforce_tls,
            smtpssl=self.smtp_enforce_ssl,
        )

        mail_sender.send(
            to=self.to,
            subject=message["Subject"],
            body=message.as_string(),
            cc=self.cc,
            _callback=kwargs.get("_callback"),
        )



from spidermon import MonitorSuite

from .monitors import (
    DownloaderExceptionMonitor,
    ErrorCountMonitor,
    FieldCoverageMonitor,
    FinishReasonMonitor,
    ItemCountMonitor,
    ItemValidationMonitor,
    PeriodicExecutionTimeMonitor,
    PeriodicItemCountMonitor,
    RetryCountMonitor,
    SuccessfulRequestsMonitor,
    TotalRequestsMonitor,
    UnwantedHTTPCodesMonitor,
    WarningCountMonitor,
)


class SpiderCloseMonitorSuite(MonitorSuite):
    """This Monitor Suite implements the following monitors:

    * :class:`.monitors.ItemCountMonitor`
    * :class:`.monitors.ItemValidationMonitor`
    * :class:`.monitors.ErrorCountMonitor`
    * :class:`.monitors.WarningCountMonitor`
    * :class:`.monitors.FinishReasonMonitor`
    * :class:`.monitors.UnwantedHTTPCodesMonitor`
    * :class:`.monitors.FieldCoverageMonitor`
    * :class:`.monitors.RetryCountMonitor`
    * :class:`.monitors.DownloaderExceptionMonitor`
    * :class:`.monitors.SuccessfulRequestsMonitor`
    * :class:`.monitors.TotalRequestsMonitor`

    You can easily enable this monitor *after* enabling Spidermon::

            SPIDERMON_SPIDER_CLOSE_MONITORS = (
                'spidermon.contrib.scrapy.monitors.SpiderCloseMonitorSuite',
            )
    """

    def __init__(
        self,
        name=None,
        monitors=None,
        monitors_finished_actions=None,
        monitors_passed_actions=None,
        monitors_failed_actions=None,
        order=None,
        crawler=None,
    ):
        super().__init__(
            name,
            monitors,
            monitors_finished_actions,
            monitors_passed_actions,
            monitors_failed_actions,
            order,
            crawler,
        )
        if dict(crawler.settings).get("SPIDERMON_MONITOR_SKIPPING_RULES"):
            skip_rules = crawler.settings.get("SPIDERMON_MONITOR_SKIPPING_RULES")
            for monitor in self.monitors:
                monitor.skip_rules = skip_rules

    monitors = [
        ItemCountMonitor,
        ItemValidationMonitor,
        ErrorCountMonitor,
        WarningCountMonitor,
        FinishReasonMonitor,
        UnwantedHTTPCodesMonitor,
        FieldCoverageMonitor,
        RetryCountMonitor,
        DownloaderExceptionMonitor,
        SuccessfulRequestsMonitor,
        TotalRequestsMonitor,
    ]


class PeriodicMonitorSuite(MonitorSuite):
    """This Monitor Suite implements the following monitors:

    * :class:`.monitors.PeriodicExecutionTimeMonitor`

    You can easily enable this monitor *after* enabling Spidermon::

            SPIDERMON_PERIODIC_MONITORS = {
                'spidermon.contrib.scrapy.monitors.PeriodicMonitorSuite': # check time in seconds,
            }
    """

    monitors = [PeriodicExecutionTimeMonitor]


class PeriodicItemCountMonitorSuite(MonitorSuite):
    """This Monitor Suite implements the following monitors:

    * :class:`.monitors.PeriodicExecutionTimeMonitor`

    You can easily enable this monitor *after* enabling Spidermon::

            SPIDERMON_PERIODIC_MONITORS = {
                'spidermon.contrib.scrapy.monitors.PeriodicItemCountMonitorSuite': # check time in seconds,
            }
    """

    monitors = [PeriodicItemCountMonitor]



import re


def calculate_field_coverage(stats):
    coverage = {}
    for key, value in stats.items():
        if not key.startswith("spidermon_item_scraped_count"):
            continue

        item_type_m = re.search(
            r"spidermon_item_scraped_count\/(?P<item_type>\w+)\/(?P<item_key>.*)", key
        )
        if item_type_m:
            item_type = item_type_m.group(1)
            item_key = item_type_m.group(2)

            item_type_total = stats.get(f"spidermon_item_scraped_count/{item_type}")

            if "/_items" in item_key:
                if item_key.endswith("/_items"):
                    continue

                levels = item_key.split("/_items/")

                root_field_type_total = stats.get(
                    f"spidermon_item_scraped_count/{item_type}/{'/_items/'.join(levels[:-1])}/_items"
                )

                item_field_coverage = value / root_field_type_total
                global_field_coverage = value / item_type_total

                coverage[
                    f"spidermon_field_coverage/{item_type}/{'/'.join(levels)}"
                ] = global_field_coverage

                coverage[
                    f"spidermon_field_coverage/{item_type}/{'/_items/'.join(levels)}"
                ] = item_field_coverage

            else:
                field_coverage = value / item_type_total

                coverage[
                    f"spidermon_field_coverage/{item_type}/{item_key}"
                ] = field_coverage

    return coverage



from . import SendSNSNotification


class SendSNSNotificationSpiderStarted(SendSNSNotification):
    def run_action(self):
        subject = "Spider Started"
        attributes = {
            "EventType": {"DataType": "String", "StringValue": "SpiderStarted"},
            "SpiderName": {"DataType": "String", "StringValue": self.data.spider.name},
            "StartTime": {
                "DataType": "String",
                "StringValue": str(self.data.stats.start_time),
            },
        }
        self.send_message(subject, attributes)


class SendSNSNotificationSpiderFinished(SendSNSNotification):
    def run_action(self):
        subject = "Spider Finished"
        attributes = {
            "EventType": {"DataType": "String", "StringValue": "SpiderFinished"},
            "SpiderName": {"DataType": "String", "StringValue": self.data.spider.name},
            "ItemsScraped": {
                "DataType": "Number",
                "StringValue": str(self.data.stats.item_scraped_count),
            },
            "FinishTime": {
                "DataType": "String",
                "StringValue": str(self.data.stats.finish_time),
            },
        }
        self.send_message(subject, attributes)



from __future__ import absolute_import

import logging

import requests
from spidermon.contrib.actions.templates import ActionWithTemplates
from spidermon.exceptions import NotConfigured

logger = logging.getLogger(__name__)


class DiscordMessageManager:
    sender_token = None

    def __init__(self, webhook_url, fake=False):
        if not webhook_url:
            raise NotConfigured(
                "You must provide a value for SPIDERMON_DISCORD_WEBHOOK_URL setting."
            )
        self.webhook_url = webhook_url
        self.fake = fake

    def send_message(self, text):
        if self.fake:
            logger.info(text)
            return

        body = {"content": text}
        response = requests.post(self.webhook_url, json=body)
        response.raise_for_status()

        if not response.ok:
            logger.error(
                f"Failed to send message. Discord API error: {response.reason}"
            )


class SendDiscordMessage(ActionWithTemplates):
    webhook_url = None
    message = None
    message_template = "discord/default/message.jinja"
    fake = False

    def __init__(
        self,
        webhook_url=None,
        message=None,
        message_template=None,
        fake=None,
    ):
        super(SendDiscordMessage, self).__init__()

        self.fake = fake or self.fake
        self.manager = DiscordMessageManager(
            webhook_url or self.webhook_url, fake=self.fake
        )
        self.message = message or self.message
        self.message_template = message_template or self.message_template

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        return {
            "webhook_url": crawler.settings.get("SPIDERMON_DISCORD_WEBHOOK_URL"),
            "message": crawler.settings.get("SPIDERMON_DISCORD_MESSAGE"),
            "message_template": crawler.settings.get(
                "SPIDERMON_DISCORD_MESSAGE_TEMPLATE"
            ),
            "fake": crawler.settings.getbool("SPIDERMON_DISCORD_FAKE"),
        }

    def run_action(self):
        self.manager.send_message(self.get_message())

    def get_message(self):
        if self.message:
            return self.render_text_template(self.message)
        else:
            return self.render_template(self.message_template)



from unittest import mock

import pytest
from scrapy.utils.test import get_crawler
from spidermon.utils import zyte


@pytest.fixture
def no_env_mock_module(monkeypatch):
    monkeypatch.delenv("SHUB_JOBKEY", raising=False)
    monkeypatch.setattr(zyte, "ScrapinghubClient", mock.MagicMock())
    return zyte


@pytest.fixture
def mock_module(monkeypatch):
    monkeypatch.setenv("SHUB_JOBKEY", "123/456/789")
    monkeypatch.setenv("SH_APIKEY", "foobar")
    monkeypatch.setattr(zyte, "ScrapinghubClient", mock.MagicMock())
    return zyte


@pytest.fixture
def settings():
    return get_crawler().settings


def test_client_creation_no_env(no_env_mock_module, settings):
    client = no_env_mock_module.Client(settings)
    assert client.available is False
    assert client.project_id is None
    assert client.spider_id is None
    assert client.job_id is None

    with pytest.raises(RuntimeError):
        client._apikey()


def test_client_creation(mock_module, settings):
    client = mock_module.Client(settings)
    assert client.available
    assert client.project_id == 123
    assert client.spider_id == 456
    assert client.job_id == 789


def test_client_property_project(mock_module, settings):
    client = mock_module.Client(settings)

    assert client._project is None
    client.project
    assert client._project is not None
    client._client.get_project.assert_called_with("123")


def test_client_property_job(mock_module, settings):
    client = mock_module.Client(settings)

    assert client._job is None
    client.job
    assert client._job is not None
    client._client.get_job.assert_called_with("123/456/789")


def test_client_property_spider(mock_module, settings):
    client = mock_module.Client(settings)
    client._job = mock.Mock()
    client._job.metadata.get.return_value = "my_awesome_spider"

    assert client._spider is None
    client.spider
    assert client._project is not None
    assert client._job is not None
    assert client._spider is not None

    client._job.metadata.get.assert_called_with("spider")
    client._project.spiders.get.assert_called_with("my_awesome_spider")


def test_client_close(mock_module, settings):
    client = mock_module.Client(settings)

    client.client
    client.close()
    client._client.close.assert_called()


@pytest.mark.parametrize("expected", [False, True])
def test_has_client(monkeypatch, expected):
    import sys
    from importlib import reload

    if not expected:
        monkeypatch.setitem(sys.modules, "scrapinghub", None)

    reload(zyte)
    assert zyte.HAS_CLIENT == expected


def test_client_settings_priority(mock_module, monkeypatch):
    shub_apikey = "SHUB_APIKEY"
    shub_jobauth = "SHUB_JOBAUTH"
    monkeypatch.setenv("SHUB_JOBAUTH", shub_jobauth)

    # Crawler settings has top priority
    settings = get_crawler(settings_dict={"SHUB_APIKEY": shub_apikey}).settings
    client = mock_module.Client(settings)
    assert client._apikey() == shub_apikey

    # then, SH_APIKEY
    empty_settings = get_crawler().settings
    client = mock_module.Client(empty_settings)
    assert client._apikey() == "foobar"

    # then,  SHUB_JOBAUTH
    monkeypatch.delenv("SH_APIKEY")
    assert client._apikey() == shub_jobauth

    # Otherwise, raise runtime error
    monkeypatch.delenv("SHUB_JOBAUTH")
    client = mock_module.Client(empty_settings)
    with pytest.raises(RuntimeError):
        client._apikey()



from .base import BaseScrapyMonitor, BaseStatMonitor
from .monitors import (
    CriticalCountMonitor,
    DownloaderExceptionMonitor,
    ErrorCountMonitor,
    FieldCoverageMonitor,
    FinishReasonMonitor,
    ItemCountMonitor,
    ItemValidationMonitor,
    PeriodicExecutionTimeMonitor,
    PeriodicItemCountMonitor,
    RetryCountMonitor,
    SuccessfulRequestsMonitor,
    TotalRequestsMonitor,
    UnwantedHTTPCodesMonitor,
    WarningCountMonitor,
    ZyteJobsComparisonMonitor,
    SPIDERMON_EXPECTED_FINISH_REASONS,
    SPIDERMON_UNWANTED_HTTP_CODES,
    SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT,
    SPIDERMON_MAX_EXECUTION_TIME,
    SPIDERMON_MAX_RETRIES,
    SPIDERMON_MIN_SUCCESSFUL_REQUESTS,
    SPIDERMON_MAX_REQUESTS_ALLOWED,
    SPIDERMON_JOBS_COMPARISON,
    SPIDERMON_JOBS_COMPARISON_STATES,
    SPIDERMON_JOBS_COMPARISON_TAGS,
    SPIDERMON_JOBS_COMPARISON_THRESHOLD,
    SPIDERMON_ITEM_COUNT_INCREASE,
)
from .suites import (
    SpiderCloseMonitorSuite,
    PeriodicMonitorSuite,
    PeriodicItemCountMonitorSuite,
)



import boto3
import logging

from spidermon.exceptions import NotConfigured
from spidermon.contrib.actions.templates import ActionWithTemplates

logger = logging.getLogger(__name__)


class SendSNSNotification(ActionWithTemplates):
    aws_access_key = None
    aws_secret_key = None
    aws_region_name = "us-east-1"
    topic_arn = None

    def __init__(
        self,
        topic_arn=None,
        aws_access_key=None,
        aws_secret_key=None,
        aws_region_name=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.topic_arn = topic_arn or self.topic_arn
        self.aws_access_key = aws_access_key or self.aws_access_key
        self.aws_secret_key = aws_secret_key or self.aws_secret_key
        self.aws_region_name = aws_region_name or self.aws_region_name

        if not self.topic_arn:
            raise NotConfigured(
                "You must provide a value for SPIDERMON_SNS_TOPIC_ARN setting."
            )
        if not self.aws_access_key:
            raise NotConfigured(
                "You must provide a value for SPIDERMON_AWS_ACCESS_KEY_ID setting."
            )
        if not self.aws_secret_key:
            raise NotConfigured(
                "You must provide a value for SPIDERMON_AWS_SECRET_ACCESS_KEY setting."
            )

    def run_action(self):
        self.send_message()

    def send_message(self, subject, attributes):
        client = boto3.client(
            service_name="sns",
            region_name=self.aws_region_name,
            aws_access_key_id=self.aws_access_key,
            aws_secret_access_key=self.aws_secret_key,
        )
        logger.info(
            f"Sending SNS message with subject: {subject} and attributes: {attributes}"
        )
        try:
            client.publish(
                TopicArn=self.topic_arn, Message=subject, MessageAttributes=attributes
            )
        except Exception as e:
            logger.error(f"Failed to send SNS message: {e}")
            raise
        logger.info(f"SNS message sent successfully!")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        return {
            "topic_arn": crawler.settings.get("SPIDERMON_SNS_TOPIC_ARN"),
            "aws_access_key": crawler.settings.get("SPIDERMON_AWS_ACCESS_KEY_ID"),
            "aws_secret_key": crawler.settings.get("SPIDERMON_AWS_SECRET_ACCESS_KEY"),
        }



import logging
import os
from collections import deque

import scrapinghub
from sh_scrapy.stats import HubStorageStatsCollector

logger = logging.getLogger(__name__)


class ScrapyCloudCollectionsStatsHistoryCollector(HubStorageStatsCollector):
    def _open_collection(self, spider):
        sh_client = scrapinghub.ScrapinghubClient()
        proj_id = os.environ.get("SCRAPY_PROJECT_ID")
        if proj_id is None:
            # not running on dash
            return

        project = sh_client.get_project(proj_id)
        collections = project.collections
        stats_location = f"{spider.name}_stats_history"
        store = collections.get_store(stats_location)
        return store

    def open_spider(self, spider):
        super().open_spider(spider)
        self.store = self._open_collection(spider)
        # note that the _open_collection method does not error if collection does not exist yet
        if self.store is None:
            return

        max_stored_stats = spider.crawler.settings.getint(
            "SPIDERMON_MAX_STORED_STATS", default=100
        )

        try:
            stats_history = [d.get("value") for d in self.store.iter()]
            stats_history = deque(stats_history, maxlen=max_stored_stats)
        except scrapinghub.client.exceptions.NotFound:
            # this happens if the stats store has not been created yet
            stats_history = deque([], maxlen=max_stored_stats)

        spider.stats_history = stats_history

    def _persist_stats(self, stats, spider):
        if self.store is None:
            return
        stats_history = spider.stats_history
        stats_history.appendleft(self._stats)
        for index, data in enumerate(stats_history):
            if index == 0:
                job_id = os.environ.get("SCRAPY_JOB", "")
                if job_id:
                    data["job_url"] = f"https://app.zyte.com/p/{job_id}"
            # this will create up to SPIDERMON_MAX_STORED_STATS objects
            # with keys 0 -> SPIDERMON_MAX_STORED_STATS - 1
            self.store.set({"_key": str(index), "value": data})



import os

try:
    from scrapinghub import ScrapinghubClient

    HAS_CLIENT = True
except ImportError:
    HAS_CLIENT = False


class Client:
    """
    Wrapper over ScrapinghubClient that indicates with its available
    attribute if it can be used to fetch data from Scrapy Cloud API.

    It expects that "SHUB_JOBKEY" be present as environment variable.
    Also, it will look up for Scrapy Cloud API key in `SHUB_APIKEY` Scrapy
    setting, then for the environment variables `SH_APIKEY` and `SHUB_JOBAUTH`.
    Note that "SHUB_JOBAUTH" can't access all API endpoints.
    """

    def __init__(self, settings):
        self.available = HAS_CLIENT and "SHUB_JOBKEY" in os.environ
        self._client = None
        self._project = None
        self._spider = None
        self._job = None
        self._settings = settings
        if self.available:
            self.job_key = os.environ["SHUB_JOBKEY"]
            self.project_id, self.spider_id, self.job_id = map(
                int, self.job_key.split("/")
            )
        else:
            self.project_id = None
            self.spider_id = None
            self.job_id = None

    @property
    def client(self):
        if not self._client:
            self._client = ScrapinghubClient(self._apikey())
        return self._client

    def _apikey(self):
        apikey = (
            self._settings.get("SHUB_APIKEY")
            or os.environ.get("SH_APIKEY")
            or os.environ.get("SHUB_JOBAUTH")
        )
        if not apikey:
            raise RuntimeError(
                "No Scrapy Cloud API key found. Please set `SHUB_APIKEY` in Scrapy settings,"
                " or either `SH_APIKEY` or `SHUB_JOBAUTH` environment variables `."
            )
        return apikey

    @property
    def project(self):
        if not self._project:
            self._project = self.client.get_project(str(self.project_id))
        return self._project

    @property
    def spider(self):
        if not self._spider:
            spider_name = self.job.metadata.get("spider")
            self._spider = self.project.spiders.get(spider_name)
        return self._spider

    @property
    def job(self):
        if not self._job:
            self._job = self.client.get_job(self.job_key)
        return self._job

    def close(self):
        if self._client:
            self._client.close()



import datetime
import json
import math
import os

from spidermon import Monitor, monitors
from spidermon.exceptions import NotConfigured
from spidermon.utils.zyte import Client
from spidermon.utils.settings import getdictorlist
from spidermon.contrib.monitors.mixins.stats import StatsMonitorMixin

from .base import BaseScrapyMonitor, BaseStatMonitor

SPIDERMON_EXPECTED_FINISH_REASONS = "SPIDERMON_EXPECTED_FINISH_REASONS"
SPIDERMON_UNWANTED_HTTP_CODES = "SPIDERMON_UNWANTED_HTTP_CODES"
SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT = "SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT"
SPIDERMON_MAX_EXECUTION_TIME = "SPIDERMON_MAX_EXECUTION_TIME"
SPIDERMON_MAX_RETRIES = "SPIDERMON_MAX_RETRIES"
SPIDERMON_MIN_SUCCESSFUL_REQUESTS = "SPIDERMON_MIN_SUCCESSFUL_REQUESTS"
SPIDERMON_MAX_REQUESTS_ALLOWED = "SPIDERMON_MAX_REQUESTS_ALLOWED"
SPIDERMON_JOBS_COMPARISON = "SPIDERMON_JOBS_COMPARISON"
SPIDERMON_JOBS_COMPARISON_STATES = "SPIDERMON_JOBS_COMPARISON_STATES"
SPIDERMON_JOBS_COMPARISON_TAGS = "SPIDERMON_JOBS_COMPARISON_TAGS"
SPIDERMON_JOBS_COMPARISON_THRESHOLD = "SPIDERMON_JOBS_COMPARISON_THRESHOLD"
SPIDERMON_ITEM_COUNT_INCREASE = "SPIDERMON_ITEM_COUNT_INCREASE"


@monitors.name("Extracted Items Monitor")
class ItemCountMonitor(BaseStatMonitor):
    """Check if spider extracted the minimum number of items.

    You can configure it using ``SPIDERMON_MIN_ITEMS`` setting.
    There's **NO** default value for this setting, if you try to use this
    monitor without setting it, it'll raise a ``NotConfigured`` exception.
    """

    stat_name = "item_scraped_count"
    threshold_setting = "SPIDERMON_MIN_ITEMS"
    assert_type = ">="


@monitors.name("Critical Count Monitor")
class CriticalCountMonitor(BaseStatMonitor):
    """Check for critical errors in the spider log.

    You can configure it using ``SPIDERMON_MAX_CRITICALS`` setting.
    There's **NO** default value for this setting, if you try to use this
    monitor without setting it, it'll raise a ``NotConfigured`` exception.

    If the job doesn't have any critical error, the monitor will be skipped."""

    stat_name = "log_count/CRITICAL"
    threshold_setting = "SPIDERMON_MAX_CRITICALS"
    assert_type = "<="
    fail_if_stat_missing = False


@monitors.name("Error Count Monitor")
class ErrorCountMonitor(BaseStatMonitor):
    """Check for errors in the spider log.

    You can configure it using ``SPIDERMON_MAX_ERRORS`` setting.
    There's **NO** default value for this setting, if you try to use this
    monitor without setting it, it'll raise a ``NotConfigured`` exception.

    If the job doesn't have any error, the monitor will be skipped."""

    stat_name = "log_count/ERROR"
    threshold_setting = "SPIDERMON_MAX_ERRORS"
    assert_type = "<="
    fail_if_stat_missing = False


@monitors.name("Warning Count Monitor")
class WarningCountMonitor(BaseStatMonitor):
    """Check for warnings in the spider log.

    You can configure it using ``SPIDERMON_MAX_WARNINGS`` setting.
    There's **NO** default value for this setting, if you try to use this
    monitor without setting it, it'll raise a ``NotConfigured`` exception.

    If the job doesn't have any warning, the monitor will be skipped."""

    stat_name = "log_count/WARNING"
    threshold_setting = "SPIDERMON_MAX_WARNINGS"
    assert_type = "<="
    fail_if_stat_missing = False


@monitors.name("Finish Reason Monitor")
class FinishReasonMonitor(BaseScrapyMonitor):
    """Check if a job has a expected finish reason.

    You can configure the expected reason with the
    ``SPIDERMON_EXPECTED_FINISH_REASONS``, it should be an ``iterable`` of
    valid finish reasons.

    The default value of this settings is: ``['finished', ]``."""

    @monitors.name("Should have the expected finished reason(s)")
    def test_should_finish_with_expected_reason(self):
        expected_reasons = self.crawler.settings.getlist(
            SPIDERMON_EXPECTED_FINISH_REASONS, ("finished",)
        )
        finished_reason = self.stats.get("finish_reason")
        msg = 'Finished with "{}" the expected reasons are {}'.format(
            finished_reason, expected_reasons
        )
        self.assertTrue(finished_reason in expected_reasons, msg=msg)


@monitors.name("Unwanted HTTP codes monitor")
class UnwantedHTTPCodesMonitor(BaseScrapyMonitor):
    """Check for maximum number of unwanted HTTP codes.
    You can configure it using ``SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT`` setting
    or ``SPIDERMON_UNWANTED_HTTP_CODES`` setting

    This monitor fails if during the spider execution, we receive
    more than the number of ``SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT``
    setting for at least one of the HTTP Status Codes in the list defined in
    ``SPIDERMON_UNWANTED_HTTP_CODES`` setting.

    Default values are:

    .. highlight:: python
    .. code-block:: python

        SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT = 10
        SPIDERMON_UNWANTED_HTTP_CODES = [400, 407, 429, 500, 502, 503, 504, 523, 540, 541]

    ``SPIDERMON_UNWANTED_HTTP_CODES`` can also be a dictionary with the HTTP Status Code
    as key and the maximum number of accepted responses with that code.

    With the following setting, the monitor will fail if more than 100 responses are
    404 errors or at least one 500 error:

    .. highlight:: python
    .. code-block:: python

        SPIDERMON_UNWANTED_HTTP_CODES = {
            400: 100,
            500: 0,
        }

    Furthermore, instead of being a numeric value, the code accepts a dictionary which can
    contain any of two keys: ``max_count`` and ``max_percentage``. The former refers to an
    absolute value and works the same way as setting an integer value. The latter refers
    to a max_percentage of the total number of requests the spider made. If both are set, the
    monitor will fail if any of the conditions are met. If none are set, it will default to
    ``DEFAULT_UNWANTED_HTTP_CODES_MAX_COUNT```.

    With the following setting, the monitor will fail if it has at least one 500 error or
    if there are more than ``min(100, 0.5 * total requests)`` 400 responses.

    .. highlight:: python
    .. code-block:: python

        SPIDERMON_UNWANTED_HTTP_CODES = {
            400: {"max_count": 100, "max_percentage": 0.5},
            500: 0,
        }

    """

    DEFAULT_UNWANTED_HTTP_CODES_MAX_COUNT = 10
    DEFAULT_UNWANTED_HTTP_CODES = [400, 407, 429, 500, 502, 503, 504, 523, 540, 541]

    @monitors.name("Should not hit the limit of unwanted http status")
    def test_check_unwanted_http_codes(self):
        unwanted_http_codes = getdictorlist(
            self.crawler,
            SPIDERMON_UNWANTED_HTTP_CODES,
            self.DEFAULT_UNWANTED_HTTP_CODES,
        )

        errors_max_count = self.crawler.settings.getint(
            SPIDERMON_UNWANTED_HTTP_CODES_MAX_COUNT,
            self.DEFAULT_UNWANTED_HTTP_CODES_MAX_COUNT,
        )

        if not isinstance(unwanted_http_codes, dict):
            unwanted_http_codes = {
                code: errors_max_count for code in unwanted_http_codes
            }

        requests = self.stats.get("downloader/request_count", 0)
        for code, max_errors in unwanted_http_codes.items():
            code = int(code)
            count = self.stats.get(f"downloader/response_status_count/{code}", 0)

            percentage_trigger = False

            if isinstance(max_errors, dict):
                absolute_max_errors = max_errors.get("max_count")
                percentual_max_errors = max_errors.get("max_percentage")

                # if the user passed an empty dict, use the default count
                if not absolute_max_errors and not percentual_max_errors:
                    max_errors = self.DEFAULT_UNWANTED_HTTP_CODES_MAX_COUNT

                else:
                    # calculate the max errors based on percentage
                    # if there's no percentage set, take the number
                    # of requests as this is the same as disabling the check
                    calculated_percentage_errors = int(
                        percentual_max_errors * requests
                        if percentual_max_errors
                        else requests
                    )

                    # takes the minimum of the two values.
                    # if no absolute max errors were set, take the number
                    # of requests as this effectively disables this check
                    max_errors = min(
                        absolute_max_errors if absolute_max_errors else requests,
                        calculated_percentage_errors,
                    )

                    # if the max errors were defined by the percentage, remember it
                    # so we can properly format the error message.
                    percentage_trigger = max_errors == calculated_percentage_errors

            stat_message = (
                "This exceeds the limit of {} ({}% of {} total requests)".format(
                    max_errors, percentual_max_errors * 100, requests
                )
                if percentage_trigger
                else "This exceeds the limit of {}".format(max_errors)
            )

            msg = (
                "Found {} Responses with status code={} - ".format(count, code)
                + stat_message
            )
            self.assertTrue(count <= max_errors, msg=msg)


@monitors.name("Downloader Exceptions monitor")
class DownloaderExceptionMonitor(BaseStatMonitor):
    """This monitor checks if the amount of downloader
    exceptions (timeouts, rejected connections, etc.) is
    lesser or equal to a specified threshold.

    This amount is provided by ``downloader/exception_count``
    value of your job statistics. If the value is not available
    in the statistics (i.e., no exception was raised), the monitor
    will be skipped.

    Configure the threshold using the ``SPIDERMON_MAX_DOWNLOADER_EXCEPTIONS``
    setting. There's **NO** default value for this setting.
    If you try to use this monitor without a value specified, a
    ``NotConfigured`` exception will be raised.
    """

    stat_name = "downloader/exception_count"
    threshold_setting = "SPIDERMON_MAX_DOWNLOADER_EXCEPTIONS"
    assert_type = "<="
    fail_if_stat_missing = False


@monitors.name("Retry Count monitor")
class RetryCountMonitor(BaseScrapyMonitor):
    """Check if any requests have reached the maximum amount of retries
    and the crawler had to drop those requests.

    You can configure it using the ``SPIDERMON_MAX_RETRIES`` setting.
    The default is ``-1`` which disables the monitor.
    """

    @monitors.name(
        "Should not hit the limit of requests that reached the maximum retry amount"
    )
    def test_maximum_retries(self):
        max_reached = self.stats.get("retry/max_reached", 0)
        threshold = self.crawler.settings.getint(SPIDERMON_MAX_RETRIES, -1)
        if threshold < 0:
            return
        msg = "Too many requests ({}) reached the maximum retry amount".format(
            max_reached
        )
        self.assertLessEqual(max_reached, threshold, msg=msg)


@monitors.name("Successful Requests monitor")
class SuccessfulRequestsMonitor(BaseScrapyMonitor):
    """Check the amount of successful requests.

    You can configure it using the ``SPIDERMON_MIN_SUCCESSFUL_REQUESTS`` setting.
    """

    @monitors.name("Should have at least the minimum number of successful requests")
    def test_minimum_successful_requests(self):
        requests = self.stats.get("downloader/response_status_count/200", 0)
        threshold = self.crawler.settings.getint(SPIDERMON_MIN_SUCCESSFUL_REQUESTS, 0)
        msg = "Too few ({}) successful requests".format(requests)
        self.assertGreaterEqual(requests, threshold, msg=msg)


@monitors.name("Total Requests monitor")
class TotalRequestsMonitor(BaseScrapyMonitor):
    """Check the total amount of requests.

    You can configure it using the ``SPIDERMON_MAX_REQUESTS_ALLOWED`` setting.
    The default is ``-1`` which disables the monitor.
    """

    @monitors.name("Should not hit the total limit of requests")
    def test_request_count_exceeded_limit(self):
        requests = self.stats.get("downloader/request_count", 0)
        threshold = self.crawler.settings.getint(SPIDERMON_MAX_REQUESTS_ALLOWED, -1)
        if threshold < 0:
            return
        msg = "Too many ({}) requests".format(requests)
        self.assertLessEqual(requests, threshold, msg=msg)


@monitors.name("Item Validation Monitor")
class ItemValidationMonitor(BaseStatMonitor):
    """This monitor checks if the amount of validation errors
    is lesser or equal to a specified threshold.

    This amount is provided by ``spidermon/validation/fields/errors``
    value of your job statistics. If the value is not available
    in the statistics (i.e., no validation errors), the monitor
    will be skipped.

    .. warning::

       You need to enable item validation in your project so
       this monitor can be used.

    Configure the threshold using the ``SPIDERMON_MAX_ITEM_VALIDATION_ERRORS``
    setting. There's **NO** default value for this setting.
    If you try to use this monitor without a value specified, a
    ``NotConfigured`` exception will be raised.
    """

    stat_name = "spidermon/validation/fields/errors"
    threshold_setting = "SPIDERMON_MAX_ITEM_VALIDATION_ERRORS"
    assert_type = "<="
    fail_if_stat_missing = False


@monitors.name("Field Coverage Monitor")
class FieldCoverageMonitor(BaseScrapyMonitor):
    """Validate if field coverage rules are met.

    To use this monitor you need to enable the ``SPIDERMON_ADD_FIELD_COVERAGE``
    setting, which will add information about field coverage to your spider
    statistics.

    To define your field coverage rules create a dictionary containing the
    expected coverage for each field you want to monitor.

    As an example, if the items you are returning from your spider are Python dictionaries
    with the following format:

    .. code-block:: python

        {
            "field_1": "some_value",
            "field_2": "some_value",
            "field_3": {
                "field_3_1": "some_value",
                "field_3_2": "some_value",
            }
        }

    A set of rules may be defined as follows:

    .. code-block:: python

        # project/settings.py
        SPIDERMON_FIELD_COVERAGE_RULES = {
            "dict/field_1": 0.4,  # Expected 40% coverage for field_1
            "dict/field_2": 1.0,  # Expected 100% coverage for field_2
            "dict/field_3": 0.8,  # Expected 80% coverage for parent field_3
            "dict/field_3/field_3_1": 0.5,  # Expected 50% coverage for nested field_3_1
        }

    You are not obligated to set rules for every field, just for the ones in which you are interested.
    Also, you can monitor nested fields if available in your returned items.

    If a field returned by your spider is a list of dicts (or objects) and you want to check their
    coverage, that is also possible. You need to set the ``SPIDERMON_LIST_FIELDS_COVERAGE_LEVELS``
    setting. This value represents for how many levels inside the list the coverage will be computed
    (if the objects inside the list also have fields that are objects/lists).
    The coverage for list fields is computed in two ways: with
    respect to the total items scraped (these values can be greater than 1) and with respect to the
    total of items in the list. The stats are in the following form:

    .. code-block:: python

        {
            "spidermon_field_coverage/dict/field2/_items/nested_field1": "some_value",
            "spidermon_field_coverage/dict/field2/nested_field1": "other_value",
        }

    The stat containing `_items` means it is calculated based on the total list items, while the
    other, based on the total number of scraped items.

    If the objects in the list also contain another list field, that coverage is also computed in
    both ways, with the total list items considered for the `_items` stat that of the innermost list.

    In case you have a job without items scraped, and you want to skip this test, you have to enable the
    ``SPIDERMON_FIELD_COVERAGE_SKIP_IF_NO_ITEM`` setting to avoid the field coverage monitor error.

    .. warning::

       Rules for nested fields will be validated against the total number of items returned.

       For the example below, rule for ``dict/field_3/field_3_1`` will validate if 50%
       of **all** items returned contains ``field_3_1``, not just the ones that contain
       parent ``field_3``.

    .. note::
       If you are returning an item type other than a dictionary, replace `dict` by the
       class name of the item you are returning.

       Considering you have an item defined as:

       .. code-block:: python

           class MyCustomItem(scrapy.Item):
               field_1 = scrapy.Field()
               field_2 = scrapy.Field()

       You must define the field coverage rules as follows:

       .. code-block:: python

           SPIDERMON_FIELD_COVERAGE_RULES = {
               "MyCustomItem/field_1": 0.4,
               "MyCustomItem/field_2": 1.0,
           }

    """

    def run(self, result):
        add_field_coverage_set = self.crawler.settings.getbool(
            "SPIDERMON_ADD_FIELD_COVERAGE", False
        )
        if not add_field_coverage_set:
            raise NotConfigured(
                "To enable field coverage monitor, set SPIDERMON_ADD_FIELD_COVERAGE=True in your project settings"
            )

        return super().run(result)

    def test_check_if_field_coverage_rules_are_met(self):
        skip_no_items = self.crawler.settings.getbool(
            "SPIDERMON_FIELD_COVERAGE_SKIP_IF_NO_ITEM", False
        )
        items_scraped = self.data.stats.get("item_scraped_count", 0)
        if skip_no_items and int(items_scraped) == 0:
            self.skipTest("No items were scraped.")

        failures = []
        field_coverage_rules = self.crawler.settings.getdict(
            "SPIDERMON_FIELD_COVERAGE_RULES"
        )
        for field, expected_coverage in field_coverage_rules.items():
            actual_coverage = self.data.stats.get(
                f"spidermon_field_coverage/{field}", 0
            )
            if actual_coverage < expected_coverage:
                failures.append(
                    "{} (expected {}, got {})".format(
                        field, expected_coverage, actual_coverage
                    )
                )

        msg = "\nThe following items did not meet field coverage rules:\n{}".format(
            "\n".join(failures)
        )
        self.assertTrue(len(failures) == 0, msg=msg)


@monitors.name("Periodic execution time monitor")
class PeriodicExecutionTimeMonitor(Monitor, StatsMonitorMixin):
    """Check for runtime exceeding a target maximum runtime.

    You can configure the maximum runtime (in seconds) using
    ``SPIDERMON_MAX_EXECUTION_TIME`` as a project setting or spider attribute."""

    @monitors.name("Maximum execution time reached")
    def test_execution_time(self):
        crawler = self.data.get("crawler")
        max_execution_time = crawler.settings.getint(SPIDERMON_MAX_EXECUTION_TIME)
        if not max_execution_time:
            return

        start_time = self.data.stats.get("start_time")
        if not start_time:
            return

        if start_time.tzinfo:
            now = self.utc_now_with_timezone()
        else:
            now = datetime.datetime.utcnow()

        duration = now - start_time

        msg = "The job has exceeded the maximum execution time"
        self.assertLess(duration.total_seconds(), max_execution_time, msg=msg)


@monitors.name("Jobs Comparison Monitor")
class ZyteJobsComparisonMonitor(BaseStatMonitor):
    """
    .. note::
       This monitor is useful when running jobs in
       `Zyte's Scrapy Cloud <https://www.zyte.com/scrapy-cloud/>`_.

    Check for a drop in scraped item count compared to previous jobs.

    You need to set the number of previous jobs to compare, using ``SPIDERMON_JOBS_COMPARISON``.
    The default is ``0`` which disables the monitor. We use the average of the scraped items count.

    You can configure which percentage of the previous item count is the minimum acceptable, by
    using the setting ``SPIDERMON_JOBS_COMPARISON_THRESHOLD``. We expect a float number between
    ``0.0`` (not inclusive) and with no upper limit (meaning we can check if itemcount is increasing
    at a certain rate). If not set, a NotConfigured error will be raised.

    You can filter which jobs to compare based on their states using the
    ``SPIDERMON_JOBS_COMPARISON_STATES`` setting. The default value is ``("finished",)``.

    You can also filter which jobs to compare based on their tags using the
    ``SPIDERMON_JOBS_COMPARISON_TAGS`` setting. Among the defined tags we consider only those
    that are also present in the current job.
    """

    stat_name = "item_scraped_count"
    assert_type = ">="

    def run(self, result):
        if (
            SPIDERMON_JOBS_COMPARISON not in self.crawler.settings.attributes
            or self.crawler.settings.getint(SPIDERMON_JOBS_COMPARISON) <= 0
        ):
            raise NotConfigured(
                f"Configure SPIDERMON_JOBS_COMPARISON to your project "
                f"settings to use {self.monitor_name}."
            )

        if (
            SPIDERMON_JOBS_COMPARISON_THRESHOLD not in self.crawler.settings.attributes
            or self.crawler.settings.getfloat(SPIDERMON_JOBS_COMPARISON_THRESHOLD) <= 0
        ):
            raise NotConfigured(
                f"Configure SPIDERMON_JOBS_COMPARISON_THRESHOLD to your project "
                f"settings to use {self.monitor_name}."
            )

        return super().run(result)

    def _get_jobs(self, states, number_of_jobs):
        tags = self._get_tags_to_filter()

        jobs = []
        start = 0
        client = Client(self.crawler.settings)

        _jobs = client.spider.jobs.list(
            start=start,
            state=states,
            count=number_of_jobs,
            filters=dict(has_tag=tags) if tags else None,
        )
        while _jobs:
            jobs.extend(_jobs)
            start += 1000
            _jobs = client.spider.jobs.list(
                start=start,
                state=states,
                count=number_of_jobs,
                filters=dict(has_tag=tags) if tags else None,
            )
        return jobs

    def _get_tags_to_filter(self):
        """
        Return the intersect of the desired tags to filter and
        the ones from the current job.
        """
        desired_tags = self.crawler.settings.getlist(SPIDERMON_JOBS_COMPARISON_TAGS)
        if not desired_tags:
            return {}

        current_tags = json.loads(os.environ.get("SHUB_JOB_DATA", "{}")).get("tags")
        if not current_tags:
            return {}

        tags_to_filter = set(desired_tags) & set(current_tags)
        return sorted(tags_to_filter)

    def get_threshold(self):
        number_of_jobs = self.crawler.settings.getint(SPIDERMON_JOBS_COMPARISON)

        threshold = self.crawler.settings.getfloat(SPIDERMON_JOBS_COMPARISON_THRESHOLD)

        states = self.crawler.settings.getlist(
            SPIDERMON_JOBS_COMPARISON_STATES, ("finished",)
        )

        jobs = self._get_jobs(states, number_of_jobs)

        previous_count = sum(job.get("items", 0) for job in jobs) / len(jobs)

        expected_item_extracted = math.ceil(previous_count * threshold)
        return expected_item_extracted


@monitors.name("Periodic Item Count Increase Monitor")
class PeriodicItemCountMonitor(BaseStatMonitor):
    """Check for increase in item count.

    You can configure the threshold for increase using
    ``SPIDERMON_ITEM_COUNT_INCREASE`` as a project setting or spider attribute.
    Use int value to check for x new items every check or float value to check
    in percentage increase of items.
    """

    stat_name = "item_scraped_count"
    threshold_setting = "SPIDERMON_ITEM_COUNT_INCREASE"
    assert_type = ">="

    def run(self, result):
        if SPIDERMON_ITEM_COUNT_INCREASE not in self.crawler.settings.attributes:
            raise NotConfigured(
                f"Configure {SPIDERMON_ITEM_COUNT_INCREASE} to your project "
                f"settings to use {self.monitor_name}."
            )

        return super().run(result)

    def get_threshold(self):
        crawler = self.data.crawler
        prev_item_scraped_count = self.stats.get("prev_item_scraped_count", 0)
        item_scraped_count = self.stats.get(self.stat_name, 0)
        crawler.stats.set_value("prev_item_scraped_count", item_scraped_count)
        threshold_increase = crawler.settings.get(self.threshold_setting)
        if isinstance(threshold_increase, int):
            return prev_item_scraped_count + threshold_increase
        elif isinstance(threshold_increase, float):
            return prev_item_scraped_count + (
                prev_item_scraped_count * threshold_increase
            )



import operator
import logging

from spidermon import Monitor
from spidermon.exceptions import NotConfigured

from ...monitors.mixins.spider import SpiderMonitorMixin

logger = logging.getLogger(__name__)


class BaseScrapyMonitor(Monitor, SpiderMonitorMixin):
    """
    Monitor can be skipped based on conditions given in the settings.
    The purpose is to skip a monitor based on stat value or any custom
    function. A scenario could be skipping the Field Coverage Monitor
    when a spider produced no items. Following is a code block of
    examples of how we can configure the skip rules in settings.

    Example #1: skip rules based on stat values
    .. code-block:: python
        class QuotesSpider(scrapy.Spider):
            name = "quotes"
            custom_settings = {
                "SPIDERMON_FIELD_COVERAGE_RULES": {
                    "dict/quote": 1,
                    "dict/author": 1,
                },
                "SPIDERMON_MONITOR_SKIPPING_RULES": {
                    "Field Coverage Monitor": [["item_scraped_count", "==", 0]],
                }
            }

    Example #2: skip rules based on a custom function
    .. code-block:: python

        def skip_function(monitor):
            return "item_scraped_count" not in monitor.data.stats

        class QuotesSpider(scrapy.Spider):
            name = "quotes"

            custom_settings = {
                "SPIDERMON_FIELD_COVERAGE_RULES": {
                    "dict/quote": 1,
                    "dict/author": 1,
                },
                "SPIDERMON_MONITOR_SKIPPING_RULES": {
                    "Field Coverage Monitor": [skip_function],
                }
            }
    """

    longMessage = False
    ops = {
        ">": operator.gt,
        ">=": operator.ge,
        "<": operator.lt,
        "<=": operator.le,
        "==": operator.eq,
        "!=": operator.ne,
    }

    @property
    def monitor_description(self):
        if self.__class__.__doc__:
            return self.__class__.__doc__.split("\n")[0]
        return super().monitor_description

    def run(self, result):
        if self.check_if_skip_rule_met():
            logger.info(f"Skipping {self.monitor_name} monitor")
            return

        return super().run(result)

    def check_if_skip_rule_met(self):
        if (
            hasattr(self, "skip_rules")
            and getattr(self, "monitor_name")
            and self.skip_rules.get(self.monitor_name)
        ):
            skip_rules = self.skip_rules[self.monitor_name]
            for rule in skip_rules:
                if hasattr(rule, "__call__"):
                    if rule(self):
                        return True
                    continue
                stats_value = self.data.stats.get(rule[0], 0)
                if self.ops[rule[1]](stats_value, rule[2]):
                    return True

        return False


class BaseStatMonitor(BaseScrapyMonitor):
    """Base Monitor class for stat-related monitors.

    Create a monitor class inheriting from this class to have a custom
    monitor that validates numerical stats from your job execution
    against a configurable threshold. If this threshold is passed in
    via command line arguments (and not it the spider settings), the setting is read as a
    string and converted to ``threshold_datatype`` type (default is
    float).

    As an example, we will create a new monitor that will check if the
    value obtained in a job stat 'numerical_job_statistic' is greater than
    or equal to the value configured in ``CUSTOM_STAT_THRESHOLD`` project
    setting:

    .. code-block:: python

        class MyCustomStatMonitor(BaseStatMonitor):
            stat_name = "numerical_job_statistic"
            threshold_setting = "CUSTOM_STAT_THRESHOLD"
            assert_type = ">="

    For the ``assert_type`` property you can select one of the following:

    ==  =====================
    >   Greater than
    >=  Greater than or equal
    <   Less than
    <=  Less than or equal
    ==  Equal
    !=  Not equal
    ==  =====================

    Sometimes, we don't want a fixed threshold, but a dynamic based on more than
    one stat or getting data external from the job execution (e.g., you want the
    threshold to be related to another stat, or you want to get the value
    of a stat from a previous job).

    As an example, the following monitor will use as threshold the a variable number
    of errors allowed based on the number of items scraped. So this monitor will pass
    only if the number of errors is less than 1% of the number of items scraped:

    .. code-block:: python

        class MyCustomStatMonitor(BaseStatMonitor):
            stat_name = "log_count/ERROR"
            assert_type = "<"

            def get_threshold(self):
                item_scraped_count = self.stats.get("item_scraped_count")
                return item_scraped_count * 0.01

    By default, if the stat can't be found in job statistics, the monitor will fail.
    If you want the monitor to be skipped in that case, you should set ``fail_if_stat_missing``
    attribute as ``False``.


    The following monitor will not fail if the job doesn't have a ``numerical_job_statistic``
    value in its statistics:

    .. code-block:: python

        class MyCustomStatMonitor(BaseStatMonitor):
            stat_name = "numerical_job_statistic"
            threshold_setting = "CUSTOM_STAT_THRESHOLD"
            assert_type = ">="
            fail_if_stat_missing = False
    """

    fail_if_stat_missing = True
    threshold_datatype = float

    @property
    def _get_threshold_setting(self):
        datatype_to_function = {
            int: self.crawler.settings.getint,
            float: self.crawler.settings.getfloat,
        }

        return datatype_to_function[self.threshold_datatype]

    def run(self, result):
        has_threshold_config = any(
            [hasattr(self, "threshold_setting"), hasattr(self, "get_threshold")]
        )
        if not has_threshold_config:
            raise NotConfigured(
                f"{self.__class__.__name__} should include a a `threshold_setting` attribute "
                "to be configured in your project settings with the desired threshold "
                "or a `get_threshold` method that returns the desired threshold."
            )

        if (
            hasattr(self, "threshold_setting")
            and self.threshold_setting not in self.crawler.settings.attributes
        ):
            raise NotConfigured(
                f"Configure {self.threshold_setting} to your project "
                f"settings to use {self.monitor_name}."
            )

        return super().run(result)

    def _get_threshold_value(self):
        if hasattr(self, "get_threshold"):
            return self.get_threshold()
        return self._get_threshold_setting(self.threshold_setting)

    def test_stat_monitor(self):
        assertions = {
            ">": self.assertGreater,
            ">=": self.assertGreaterEqual,
            "<": self.assertLess,
            "<=": self.assertLessEqual,
            "==": self.assertEqual,
            "!=": self.assertNotEqual,
        }
        threshold = self._get_threshold_value()

        if self.stat_name not in self.stats:
            message = f"Unable to find '{self.stat_name}' in job stats."
            if self.fail_if_stat_missing:
                self.fail(message)
            else:
                self.skipTest(message)

        value = self.stats.get(self.stat_name)

        assertion_method = assertions.get(self.assert_type)
        assertion_method(
            value,
            threshold,
            msg=f"Expecting '{self.stat_name}' to be '{self.assert_type}' "
            f"to '{threshold}'. Current value: '{value}'",
        )



from spidermon.core.actions import Action
from spidermon.exceptions import NotConfigured, SkipAction
from unittest.mock import MagicMock


def test_action_success():
    class TestAction(Action):
        def run_action(self):
            pass

    result_mock = MagicMock()
    action = TestAction()

    action.run(result_mock, MagicMock())

    result_mock.add_action_success.assert_called()


def test_action_fail():
    class TestAction(Action):
        def run_action(self):
            raise Exception

    result_mock = MagicMock()
    action = TestAction()

    action.run(result_mock, MagicMock())

    result_mock.add_action_error.assert_called()
    result_mock.add_action_success.assert_not_called()


def test_action_skip():
    class TestAction(Action):
        def run_action(self):
            raise SkipAction("Test")

    result_mock = MagicMock()
    action = TestAction()

    action.run(result_mock, MagicMock())

    result_mock.add_action_skip.assert_called()


def test_fallback_action():
    fallback_mock = MagicMock()

    class TestAction(Action):
        fallback = fallback_mock

        def run_action(self):
            raise Exception

    action = TestAction()
    action.run(MagicMock(), MagicMock())

    fallback_mock.assert_called()
    fallback_mock().run.assert_called()


def test_fallback_skip_action():
    # fallback not called for SkipAction exception
    fallback_mock = MagicMock()

    class TestAction(Action):
        fallback = fallback_mock

        def run_action(self):
            raise SkipAction("Test")

    action = TestAction()

    action.run(MagicMock(), MagicMock())
    fallback_mock().run.assert_not_called()


def test_fallback_not_configured():
    # raises not configured error for unconfigured fallback actions
    fallback_mock = MagicMock()
    fallback_mock.side_effect = NotConfigured

    class TestAction(Action):
        fallback = fallback_mock

        def run_action(self):
            pass

    try:
        TestAction()
    except NotConfigured:
        assert True
    else:
        assert False



from typing import Any, List
from itemadapter import ItemAdapter


def traverse_nested(obj: ItemAdapter, keys: List[str]) -> ItemAdapter:
    """
    Get the last nested attribute from a list of keys within an ItemAdapter object.
    Raises:
        KeyError: if any of the keys in the path is not defined.
    """
    current_obj = obj
    while keys:
        try:
            # Traverse next level of item object
            key = keys.pop(0)
            current_obj = ItemAdapter(current_obj[key])
        except KeyError:
            raise KeyError(f'Invalid key "{key}" for {current_obj} in {obj}')

    return current_obj


def get_nested_attribute(item: ItemAdapter, attribute_path: str):
    """
    Get the value of a nested attribute within an ItemAdapter.

    Raises:
        KeyError: if any of the keys in the path is not defined.
    """
    *keys, last_key = attribute_path.split(".")
    nested_obj = traverse_nested(item, keys)
    return nested_obj.get(last_key)


def set_nested_attribute(item: ItemAdapter, attribute_path: str, value: Any):
    """
    Set the value of a nested attribute within an ItemAdapter.

    Raises:
        KeyError:  if any of the keys in the path is not defined or
            if the last key in the path is not supported by its parent field.
    """
    *keys, last_key = attribute_path.split(".")
    nested_obj = traverse_nested(item, keys)
    if not isinstance(nested_obj, ItemAdapter):
        nested_obj = ItemAdapter(nested_obj)

    nested_obj[last_key] = value



from setuptools import find_packages, setup

setup(
    name="spidermon",
    version="1.21.0",
    url="https://github.com/scrapinghub/spidermon",
    author="Zyte",
    author_email="opensource@zyte.com",
    description=("Spidermon is a framework to build monitors for Scrapy spiders."),
    long_description=("Spidermon is a framework to build monitors for Scrapy spiders."),
    license="BSD",
    packages=find_packages(),
    package_data={"spidermon": ["VERSION"]},
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        "jsonschema[format]>=4.21.0",
        "python-slugify",
    ],
    extras_require={
        # Specific monitors and tools to support notifications and reports
        "monitoring": [
            "Jinja2",
            "boto",
            "boto3",
            "itemadapter",
            "premailer",
            "requests",
            "scrapinghub",
            "scrapinghub-entrypoint-scrapy",
            "scrapy",
            "sentry_sdk",
            "slack_sdk",
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Scrapy",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: System :: Monitoring",
    ],
    python_requires=">=3.8",
)



import pytest
from scrapy import Item, signals
from scrapy.spiders import Spider
from scrapy.utils.test import get_crawler


class TestItem(Item):
    __test__ = False


@pytest.fixture
def spider():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
    }
    crawler = get_crawler(settings_dict=settings)

    spider = Spider.from_crawler(crawler, "example.com")

    return spider


def test_add_stats_item_scraped_count_by_item_type(spider):
    for _ in range(15):
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item={"_type": "regular_dict"},
            response="",
            spider=spider,
        )

    for _ in range(20):
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=Item(),
            response="",
            spider=spider,
        )

    for _ in range(25):
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=TestItem(),
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count") == 60
    assert stats.get("spidermon_item_scraped_count/dict") == 15
    assert stats.get("spidermon_item_scraped_count/Item") == 20
    assert stats.get("spidermon_item_scraped_count/TestItem") == 25


def test_item_scraped_count_single_field(spider):
    returned_items = [{"field1": "value1"}]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 1


def test_item_scraped_count_multiple_field(spider):
    returned_items = [{"field1": "value1", "field2": "value2"}]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 1


def test_item_scraped_count_multiple_items(spider):
    returned_items = [
        {"field1": "value1", "field2": "value2"},
        {"field1": "value1", "field2": "value2"},
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2


def test_item_scraped_count_multiple_items_field_missing(spider):
    returned_items = [
        {"field1": "value1", "field2": "value2"},
        {
            "field1": "value1",
        },
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 1


def test_item_scraped_count_single_nested_field(spider):
    returned_items = [{"field1": {"field1.1": "value1.1"}}]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.1") == 1


def test_item_scraped_count_multiple_nested_field(spider):
    returned_items = [
        {
            "field1": {"field1.1": "value1.1"},
            "field2": "value2",
            "field3": {"field3.1": "value3.1"},
            "field4": {
                "field4.1": {
                    "field4.1.1": "value",
                    "field4.1.2": "value",
                    "field4.1.3": {"field4.1.3.1": "value"},
                }
            },
        },
        {
            "field1": {
                "field1.1": "value1.1",
                "field1.2": "value1.2",
            },
            "field2": "value2",
        },
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.2") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field3") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field4") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field4/field4.1") == 1
    assert (
        stats.get("spidermon_item_scraped_count/dict/field4/field4.1/field4.1.1") == 1
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field4/field4.1/field4.1.2") == 1
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field4/field4.1/field4.1.3") == 1
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field4/field4.1/field4.1.3/field4.1.3.1"
        )
        == 1
    )


def test_item_scraped_count_multiple_nested_field_with_limit():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
        "SPIDERMON_DICT_FIELDS_COVERAGE_LEVELS": 1,
    }
    returned_items = [
        {
            "field1": {"field1.1": "value1.1"},
            "field2": "value2",
            "field3": {"field3.1": "value3.1"},
            "field4": {
                "field4.1": {
                    "field4.1.1": "value",
                    "field4.1.2": "value",
                    "field4.1.3": {"field4.1.3.1": "value"},
                }
            },
        },
        {
            "field1": {
                "field1.1": "value1.1",
                "field1.2": "value1.2",
            },
            "field2": "value2",
        },
    ]
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")
    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.2") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field3") == 1
    assert (
        stats.get("spidermon_item_scraped_count/dict/field4/field4.1/field4.1.1")
        == None
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field4/field4.1/field4.1.2")
        == None
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field4/field4.1/field4.1.3/field4.1.3.1"
        )
        == None
    )


def test_item_scraped_count_multiple_nested_field_with_two_levels_limit():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
        "SPIDERMON_DICT_FIELDS_COVERAGE_LEVELS": 2,
    }
    returned_items = [
        {
            "field1": {"field1.1": "value1.1"},
            "field2": "value2",
            "field3": {"field3.1": "value3.1"},
            "field4": {
                "field4.1": {
                    "field4.1.1": "value",
                    "field4.1.2": "value",
                    "field4.1.3": {"field4.1.3.1": "value"},
                }
            },
        },
        {
            "field1": {
                "field1.1": "value1.1",
                "field1.2": "value1.2",
            },
            "field2": "value2",
        },
    ]
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")
    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.2") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field3") == 1
    assert (
        stats.get("spidermon_item_scraped_count/dict/field4/field4.1/field4.1.1") == 1
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field4/field4.1/field4.1.2") == 1
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field4/field4.1/field4.1.3/field4.1.3.1"
        )
        == None
    )


def test_item_scraped_count_multiple_nested_field_with_no_nested_levels():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
        "SPIDERMON_DICT_FIELDS_COVERAGE_LEVELS": 0,
    }
    returned_items = [
        {
            "field1": {"field1.1": "value1.1"},
            "field2": "value2",
            "field3": {"field3.1": "value3.1"},
            "field4": {
                "field4.1": {
                    "field4.1.1": "value",
                    "field4.1.2": "value",
                    "field4.1.3": {"field4.1.3.1": "value"},
                }
            },
        },
        {
            "field1": {
                "field1.1": "value1.1",
                "field1.2": "value1.2",
            },
            "field2": "value2",
        },
    ]
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")
    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.1") == None
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.2") == None
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field3") == 1
    assert (
        stats.get("spidermon_item_scraped_count/dict/field4/field4.1/field4.1.1")
        == None
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field4/field4.1/field4.1.2")
        == None
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field4/field4.1/field4.1.3/field4.1.3.1"
        )
        == None
    )


def test_do_not_add_field_coverage_when_spider_closes_if_do_not_have_field_coverage_settings():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": False,
    }
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")

    item = {"field1": "value1"}
    spider.crawler.signals.send_catch_log_deferred(
        signal=signals.item_scraped,
        item=item,
        response="",
        spider=spider,
    )  # Return item to have some stats to calculate coverage

    crawler.signals.send_catch_log(
        signal=signals.spider_closed, spider=spider, reason=None
    )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_field_coverage/dict/field1") is None


def test_add_field_coverage_when_spider_closes_if_have_field_coverage_settings():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
    }
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")

    item = {"field1": "value1"}
    spider.crawler.signals.send_catch_log_deferred(
        signal=signals.item_scraped,
        item=item,
        response="",
        spider=spider,
    )  # Return item to have some stats to calculate coverage

    crawler.signals.send_catch_log(
        signal=signals.spider_closed, spider=spider, reason=None
    )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_field_coverage/dict/field1") == 1.0


def test_item_scraped_count_ignore_none_values():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
        "SPIDERMON_FIELD_COVERAGE_SKIP_NONE": True,
    }

    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")

    returned_items = [
        {"field1": "value1", "field2": "value2"},
        {"field1": "value1", "field2": None},
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 1


def test_item_scraped_count_do_not_ignore_none_values():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
        "SPIDERMON_FIELD_COVERAGE_SKIP_NONE": False,
    }
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")

    returned_items = [
        {"field1": "value1", "field2": "value2"},
        {"field1": "value1", "field2": None},
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2


def test_item_scraped_count_do_not_ignore_none_values_by_default(spider):
    returned_items = [
        {"field1": "value1", "field2": "value2"},
        {"field1": "value1", "field2": None},
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2


def test_item_scraped_count_list_of_dicts_disabled(spider):
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
        "SPIDERMON_LIST_FIELDS_COVERAGE_LEVELS": 0,
    }
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")
    returned_items = [
        {
            "field1": 1,
            "field2": [
                {
                    "nested_field1": 1,
                    "nested_field2": 1,
                    "nested_field3": [
                        {"deep_field1": 1},
                        {"deep_field1": 1},
                        {"deep_field2": 1},
                    ],
                },
                {"nested_field2": 1},
            ],
        },
        {
            "field1": 1,
            "field2": [
                {"nested_field1": 1},
                {
                    "nested_field1": 1,
                    "nested_field4": {"deep_field1": 1, "deep_field2": 1},
                },
                {"nested_field1": 1, "nested_field2": 1},
            ],
        },
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2

    assert stats.get("spidermon_item_scraped_count/dict/field2/_items") == None
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field1")
        == None
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field2")
        == None
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field3")
        == None
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items"
        )
        == None
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items/deep_field1"
        )
        == None
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items/deep_field2"
        )
        == None
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field4")
        == None
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field4/deep_field1"
        )
        == None
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field4/deep_field2"
        )
        == None
    )


def test_item_scraped_count_list_of_dicts_one_nesting_level(spider):
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
        "SPIDERMON_LIST_FIELDS_COVERAGE_LEVELS": 1,
    }
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")
    returned_items = [
        {
            "field1": 1,
            "field2": [
                {
                    "nested_field1": 1,
                    "nested_field2": 1,
                    "nested_field3": [
                        {"deep_field1": 1},
                        {"deep_field1": 1},
                        {"deep_field2": 1},
                    ],
                },
                {"nested_field2": 1},
            ],
        },
        {
            "field1": 1,
            "field2": [
                {"nested_field1": 1},
                {
                    "nested_field1": 1,
                    "nested_field4": {"deep_field1": 1, "deep_field2": 1},
                },
                {"nested_field1": 1, "nested_field2": 1},
            ],
        },
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2

    assert stats.get("spidermon_item_scraped_count/dict/field2/_items") == 5
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field1") == 4
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field2") == 3
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field3") == 1
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items"
        )
        == None
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items/deep_field1"
        )
        == None
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items/deep_field2"
        )
        == None
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field4") == 1
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field4/deep_field1"
        )
        == 1
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field4/deep_field2"
        )
        == 1
    )


def test_item_scraped_count_list_of_dicts_two_nesting_levels(spider):
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
        "SPIDERMON_LIST_FIELDS_COVERAGE_LEVELS": 2,
    }
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")
    returned_items = [
        {
            "field1": 1,
            "field2": [
                {
                    "nested_field1": 1,
                    "nested_field2": 1,
                    "nested_field3": [
                        {"deep_field1": 1},
                        {"deep_field1": 1},
                        {"deep_field2": 1},
                    ],
                },
                {"nested_field2": 1},
            ],
        },
        {
            "field1": 1,
            "field2": [
                {"nested_field1": 1},
                {
                    "nested_field1": 1,
                    "nested_field4": {"deep_field1": 1, "deep_field2": 1},
                },
                {"nested_field1": 1, "nested_field2": 1},
            ],
        },
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2

    assert stats.get("spidermon_item_scraped_count/dict/field2/_items") == 5
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field1") == 4
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field2") == 3
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field3") == 1
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items"
        )
        == 3
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items/deep_field1"
        )
        == 2
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items/deep_field2"
        )
        == 1
    )
    assert (
        stats.get("spidermon_item_scraped_count/dict/field2/_items/nested_field4") == 1
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field4/deep_field1"
        )
        == 1
    )
    assert (
        stats.get(
            "spidermon_item_scraped_count/dict/field2/_items/nested_field4/deep_field2"
        )
        == 1
    )



from spidermon.utils.field_coverage import calculate_field_coverage


def test_calculate_field_coverage_from_stats():
    spider_stats = {
        "finish_reason": "finished",
        "spidermon_item_scraped_count": 100,
        "spidermon_item_scraped_count/dict": 100,
        "spidermon_item_scraped_count/dict/author": 100,
        "spidermon_item_scraped_count/dict/author/author_url": 64,
        "spidermon_item_scraped_count/dict/author/name": 100,
        "spidermon_item_scraped_count/dict/quote": 50,
        "spidermon_item_scraped_count/dict/tags": 100,
    }

    expected_coverage = {
        "spidermon_field_coverage/dict/author": 1.0,
        "spidermon_field_coverage/dict/author/author_url": 0.64,
        "spidermon_field_coverage/dict/author/name": 1.0,
        "spidermon_field_coverage/dict/quote": 0.5,
        "spidermon_field_coverage/dict/tags": 1.0,
    }

    coverage = calculate_field_coverage(spider_stats)

    assert coverage == expected_coverage


def test_calculate_field_coverage_from_stats_with_nested_fields():
    spider_stats = {
        "finish_reason": "finished",
        "spidermon_item_scraped_count": 100,
        "spidermon_item_scraped_count/dict": 100,
        "spidermon_item_scraped_count/dict/field1": 100,
        "spidermon_item_scraped_count/dict/field2": 90,
        "spidermon_item_scraped_count/dict/field2/_items": 1000,
        "spidermon_item_scraped_count/dict/field2/_items/nested_field1": 550,
        "spidermon_item_scraped_count/dict/field2/_items/nested_field2": 1000,
        "spidermon_item_scraped_count/dict/field2/_items/nested_field3": 300,
        "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items": 500,
        "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items/deep_field1": 500,
        "spidermon_item_scraped_count/dict/field2/_items/nested_field3/_items/deep_field2": 250,
        "spidermon_item_scraped_count/dict/field2/_items/nested_field4": 500,
        "spidermon_item_scraped_count/dict/field2/_items/nested_field4/deep_field1": 500,
        "spidermon_item_scraped_count/dict/field2/_items/nested_field4/deep_field2": 250,
    }

    expected_coverage = {
        "spidermon_field_coverage/dict/field1": 1.0,
        "spidermon_field_coverage/dict/field2": 0.9,
        "spidermon_field_coverage/dict/field2/_items/nested_field1": 0.55,
        "spidermon_field_coverage/dict/field2/_items/nested_field2": 1.0,
        "spidermon_field_coverage/dict/field2/_items/nested_field3": 0.3,
        "spidermon_field_coverage/dict/field2/_items/nested_field3/_items/deep_field1": 1.0,
        "spidermon_field_coverage/dict/field2/_items/nested_field3/_items/deep_field2": 0.5,
        "spidermon_field_coverage/dict/field2/_items/nested_field4": 0.5,
        "spidermon_field_coverage/dict/field2/_items/nested_field4/deep_field1": 0.5,
        "spidermon_field_coverage/dict/field2/_items/nested_field4/deep_field2": 0.25,
        "spidermon_field_coverage/dict/field2/nested_field1": 5.5,
        "spidermon_field_coverage/dict/field2/nested_field2": 10.0,
        "spidermon_field_coverage/dict/field2/nested_field3": 3.0,
        "spidermon_field_coverage/dict/field2/nested_field3/deep_field1": 5.0,
        "spidermon_field_coverage/dict/field2/nested_field3/deep_field2": 2.5,
        "spidermon_field_coverage/dict/field2/nested_field4": 5.0,
        "spidermon_field_coverage/dict/field2/nested_field4/deep_field1": 5.0,
        "spidermon_field_coverage/dict/field2/nested_field4/deep_field2": 2.5,
    }

    coverage = calculate_field_coverage(spider_stats)

    assert coverage == expected_coverage


def test_calculate_field_coverage_field_ends_with_items():
    spider_stats = {
        "finish_reason": "finished",
        "spidermon_item_scraped_count": 100,
        "spidermon_item_scraped_count/dict": 100,
        "spidermon_item_scraped_count/dict/field_ends_with_items": 100,
    }

    expected_coverage = {"spidermon_field_coverage/dict/field_ends_with_items": 1.0}

    coverage = calculate_field_coverage(spider_stats)
    assert coverage == expected_coverage
